# Light Fogging Correction tool for ImageJ
# Ver. 0.028
# (c)2021 Ohnishi, Yasuo
from ij import IJ, ImagePlus
from ij.io import DirectoryChooser, OpenDialog
from ij.gui import GenericDialog
from ij.plugin import ChannelSplitter, Duplicator
from ij.plugin import RGBStackMerge, RGBStackConverter
from ij.process import ImageConverter, ImageProcessor
from loci.plugins import BF
import math
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')
import datetime

def convValue(Value, Mode):
#   Converting Luminance Value each in 8, 16, 32bit mode
    if Mode == "16-bit":
        Value = Value * 256
    elif Mode == "32-bit":
        Value = Value * 256 * 256
    else:
        Value = Value
    return Value

####### calculating R-G mean difference routine ############
def calcMean(ip0, ip1, xStart, xEnd, nYDivide, yStart, yEnd):
#   ip0: R image / ip1: G image / xStart & xEnd: Starting & Ending Point (x) of ROI
#   nYDivide:  number of dividing zones in Y
    impTmp = []
    impTmp.append(ip0)
    impTmp.append(ip1)
    nYDivide = int(nYDivide) # number of dividing zones in Y
    width = ip0.getWidth() # get image width
    height = ip0.getHeight() # get image height
    pixelNum = width*height # get total pixel number
    print ("W x H (ip0)", str(width), str(height))
    if yStart == 0 and yEnd == 0:
        height2 = height
    else:
        height2 = yEnd - yStart
    zoneBreadth = height2 // nYDivide # divided Y zone height
    tmpsumDiff = [0] * nYDivide # sum of R-G differance in a zone
    n_tmpsumDiff = [0] * nYDivide # sum of pixel number in a zone
    mean2 = [0.0] * nYDivide # mean R-G differance in a zone

    for x in range(xStart, xEnd): # processing in each X row
#       ----   
        for yz in range(nYDivide): # processing in each Y zone
            for yj in range(yStart + yz * zoneBreadth, yStart + (yz + 1) * zoneBreadth - 1): 
                Px0 = impTmp[0].getPixel(x, yj) # get R pixel Value
                Px1 = impTmp[1].getPixel(x, yj) # get G pixel Value
                tmpsumDiff[yz] = tmpsumDiff[yz] + (Px0[0] - Px1[0]) # summate R-G difference in each Y Zone
                n_tmpsumDiff[yz] = n_tmpsumDiff[yz] + 1 # count pixel number in each Y Zone
#       ----   
    for yz in range(nYDivide): # calculate mean R-G difference in each Y Zone
        mean2[yz] = float(tmpsumDiff[yz]) / float(n_tmpsumDiff[yz])
#       ----   
    return mean2 # Return value: mean R-G difference in Y Zones



######### Image correcting routine ################
def image_correct(impR, impR2, impG, xStart, xEnd, Mode, GmixRatio, RsubRatio, nYDivide, mean2, yStart, yEnd): 
# impR: R Image impR2: R image to be corrected mean: mean differance of R-G
# xStart/xEnd: x start/end value of ROI, Mode: bit depth of image: GmixRatio: Maximum R Ch. mixing ratio
# RsubRatio: Maximum Rch. value subtracting ratio nYDivide: dividing number of ROI in Y direction
# mean2: mean R-G differance value in each divided ROIs 
    nYDivide = int(nYDivide) # number of Y zone
    width = impR.getWidth() # get image width
    if xEnd > width:
        xEnd = width
    height = impG.getHeight() # get image height
    if yStart == 0 and yEnd == 0:
        height2 = height
    else:
        height2 = yEnd - yStart
    correctValue = []
    maxcorrectValue = 0
    maxCorV = [0.0 for yz in range(nYDivide)]
#### Getting R-G value difference in each pixel
    tmpMxDiffRG = [[0.0 for yj in range(height2)] for xi in range(xEnd - xStart)]
    timenow = datetime.datetime.now()
    print("Start Read tmpMxDiffRG:", timenow)
    for x in range(xStart, xEnd): # process in each x col
        for y in range(yStart, yStart + height2):
#                   get Red image pixel value
            PxR = impR.getPixel(x, y) 
#                   get Green image pixel value
            PxG = impG.getPixel(x, y) 
            tmpMxDiffRG[x - xStart][y - yStart] = PxR[0] - PxG[0]
                 # save R-G pixel value difference to tmpMxDiffRG
#------------------x, y roop end
    correctValueMx = [[0.0 for yz in range(nYDivide)] for xi in range(xEnd - xStart)]
    zoneBreadth = height2 // nYDivide # height of a Y zone
    print ("zoneBreadth: ", zoneBreadth)
    zoneLUT = [0] * height2 # set look up table for y -> yzone
    for y in range(yStart, yStart + height2):
        zoneLUT[y] = min (y // zoneBreadth, nYDivide - 1) # make LUT for y -> yzone
    timenow = datetime.datetime.now()
    print("Start Calculate correctValueMx:", timenow)
#### Calculating correcting value
    for xi in range(xEnd - xStart): # process in each x row in ROI
        for yz in range(nYDivide):
            sumtmpDiffRG = 0.0
            nr = 0.0
            if (xi == 0) and (yz == int(nYDivide /2)):
                timenow = datetime.datetime.now()
                print("Calcurate ", xi, "th correctValueMx:", timenow)
            for xi2 in range(xi-2, xi+2): # calculating back and forth 5 rows
                for yj in range(yStart + yz * zoneBreadth, yStart + (yz + 1) * zoneBreadth - 1): 
                    if (xi2 > 0) and (yj > 0) and (xi2 < (xEnd - xStart)) and (yj < yStart + height2): 
                       # summate R-G value difference in each Y zone
                        nr = nr + 1.0
                        sumtmpDiffRG = sumtmpDiffRG + tmpMxDiffRG[xi2][yj - yStart] #
            if nr > 0.0:
                correctValueMx[xi][yz] = (sumtmpDiffRG / nr) - mean2[yz]
#       define correcting value
#           divide sum of R-G diff with pixel number and subtract normal R-G diff mean
            else: 
                correctValueMx[xi][yz] = 0.0
            if correctValueMx[xi][yz] < 0.0: # if R mean < G mean in a Y zone then no correction
                correctValueMx[xi][yz] = 0.0 
            if (xi % 100 == 0) and (yz == 5):
                 print ("xi: ", xi, "yz: ", yz, "correctValueMx: ", correctValueMx[xi][yz])
            maxcorrectValue = max(maxcorrectValue, correctValueMx[xi][yz]) # calculate max correction value
            maxCorV[yz] = max(maxCorV[yz], correctValueMx[xi][yz]) # calculate max correction value
    print ("maxcorrectValue", maxcorrectValue)
### value correction process
    timenow = datetime.datetime.now()
    print("Start Correction:", timenow)
    for x in range(xStart, xEnd): # in each x row in ROI
        for y in range(yStart, yStart + height2): # start R value correction
            PxR= impR.getPixel(x, y)
            PxG= impG.getPixel(x, y)
            if correctValueMx[x - xStart][zoneLUT[y - yStart]] < 0.0: # if correction value < 0 then no correction
                correctValueMx[x - xStart][zoneLUT[y - yStart]] = 0.0
            Pcc0 = abs(correctValueMx[x - xStart][zoneLUT[y - yStart]]) / maxcorrectValue
#            Pcc0 = abs(correctValueMx[x - xStart][zoneLUT[y - yStart]]) / maxCorV[zoneLUT[y]]
#             calculate correction value ratio
            Pcc1 = Pcc0 ** (1/2.0) # gamma weighted
            Pcc = Pcc1 * GmixRatio # GmixRatio: Max G ch. mixing Ratio (like... 0.7 or 0.9)
#             define corrected pixel value in R
#            Pxc = PxG[0] * Pcc + (PxR[0] - convValue(256, Mode) * RsubRatio * Pcc1 * (PxR[0]/convValue(256, Mode))) * (1-Pcc)
#            Pxc = (PxG[0] * Pcc + PxR[0] * (1-Pcc) ) * (1- RsubRatio * Pcc0) 
            Pxc = (PxG[0] * Pcc + PxR[0] * (1-Pcc) * (1- RsubRatio * Pcc0) ) 
            if (x % 100 == 0) and (y == int(height2)/5 + yStart):
                HPcc = Pcc
            elif (x % 100 == 0) and (y == int(height2)/2 + yStart):
                MPcc = Pcc
            elif (x % 100 == 0) and (y == 4 * int(height2)/5 + yStart):
                LPcc = Pcc
                print("x(",x,")  G Mixing Ratio L:", LPcc, "M: ",MPcc, "H: ",HPcc)
            if Pxc <= 0:
                Pxc = 0
            elif Pxc >= (256 * convValue(1, Mode) - 1):
                Pxc = (256 * convValue(1, Mode) - 1)
            impR2.getProcessor().set(x, y, int(Pxc))

### Get parameters from the input DialogBox 
def getParameters():
    listRSE = [[0 for i in range(2)] for j in range(3)]
    gd = GenericDialog("Light Fog Correction Tool: (c)2021 Yasuo Ohnishi")
    gd.addMessage("Please Enter Parameters")
    mode = ["Auto", "16-bit","8-bit"]
    gd.addChoice("Image bit depth", mode, mode[0])
    gd.addNumericField("Max Mixing Ratio of G Value (0.0 to 1.0)", 1.00, 2)
    gd.addNumericField("Max Reduce Ratio of R Value (0.0 to 1.0)", 0.10, 2)
    gd.addMessage("1st Region to be corrected")
    gd.addNumericField("Start Point (x)", 0, 0)
    gd.addNumericField("End Point (x)", 0, 0)
    gd.addMessage("2nd Region to be corrected")
    gd.addNumericField("Start Point (x)", 0, 0)
    gd.addNumericField("End Point (x)", 0, 0)
    gd.addMessage("3rd Region to be corrected")
    gd.addNumericField("Start Point (x)", 0, 0)
    gd.addNumericField("End Point (x)", 0, 0)
    gd.addMessage("Vertical Dividing Zone")
    gd.addNumericField("Number of Y Dividing Zone", 10, 0)
    gd.addMessage("Define Sampling Normal Region (Optional)")
    gd.addNumericField("Start Point (x)", 0, 0)
    gd.addNumericField("End Point (x)", 0, 0)
    Show = ["Yes","No"]
    gd.addChoice("Display working windows", Show, Show[0])
    YresAlt = ["Yes","No"]
    gd.addChoice("Restrict Correcting Y Zone", YresAlt, YresAlt[1])
    gd.showDialog()
#
    parm_Mode   = gd.getNextChoice()
    GmixRatio = gd.getNextNumber()
    RsubRatio = gd.getNextNumber()
    listRSE[0][0] =  int(gd.getNextNumber()) # 1st start
    listRSE[0][1] =  int(gd.getNextNumber()) # 1st end
    listRSE[1][0] =  int(gd.getNextNumber()) # 2ndt start
    listRSE[1][1] =  int(gd.getNextNumber()) # 2nd end
    listRSE[2][0] =  int(gd.getNextNumber()) # 3rd start
    listRSE[2][1] =  int(gd.getNextNumber()) # 3rd end
    nYDivide = gd.getNextNumber()
    SmplXStart = gd.getNextNumber()
    SmplXEnd = gd.getNextNumber()
    parm_Show   = gd.getNextChoice()
    Yres   = gd.getNextChoice()
    if GmixRatio > 1.0:
        GmixRatio = 1.0
    if RsubRatio > 1.0:
        RsubRatio = 1.0
#
    if gd.wasCanceled():  
        parm_Mode   = "Stop"
        IJ.log("Canceled")
    if Yres == "Yes":  
        IJ.log("Yres: Yes")

#
    return str(parm_Mode), str(parm_Show), listRSE, GmixRatio, RsubRatio, nYDivide, int(SmplXStart), int(SmplXEnd), Yres


### 2nd / Get parameters from the input DialogBox 
def getParameters2(RegionSE, nYDivide, SmplXStart, SmplXEnd):
    listRSE = [[0 for i in range(2)] for j in range(3)]
    gd = GenericDialog("Light Fog Correction Tool: (c)2021 Yasuo Ohnishi")
    gd.addMessage("Please Enter Parameters")
    gd.addMessage("Region to be corrected")
    gd.addNumericField("Start Point (x)", RegionSE[0][0], 0)
    gd.addNumericField("End Point (x)", RegionSE[0][1], 0)
    gd.addNumericField("Start Point (y)", 0, 0)
    gd.addNumericField("End Point (y)", 0, 0)
    gd.addMessage("Vertical Dividing Zone")
    gd.addNumericField("Number of Y Dividing Zone", nYDivide, 0)
    gd.addMessage("Define Sampling Normal Region (Optional)")
    gd.addNumericField("Start Point (x)", SmplXStart, 0)
    gd.addNumericField("End Point (x)", SmplXEnd, 0)
    gd.showDialog()
#
    RegionSE[0][0] =  int(gd.getNextNumber()) # 1st start
    RegionSE[0][1] =  int(gd.getNextNumber()) # 1st end
    yStart =  int(gd.getNextNumber()) # 1st start
    yEnd =  int(gd.getNextNumber()) # 1st end
    nYDivide = gd.getNextNumber()
    SmplXStart = gd.getNextNumber()
    SmplXEnd = gd.getNextNumber()
#
    if gd.wasCanceled():  
        parm_Mode   = "Stop"
        IJ.log("Canceled")
    if Yres == "Yes":  
        IJ.log("Yres: Yes")
    RegionSE[1][0] =  0 # 2ndt start
    RegionSE[1][1] =  0 # 2nd end
    RegionSE[2][0] =  0 # 3rd start
    RegionSE[2][1] =  0 # 3rd end

#
    return RegionSE, int(nYDivide), int(SmplXStart), int(SmplXEnd), int(yStart), int(yEnd)





#----------------------------------------------
def manipulation(Mode, Show, RegionSE, folder, filename, BmixRatio, RsubRatio, nYDivide, SmplXStart, SmplXEnd, yStart, yEnd):
# Always Open file with Bio-format Plugin
    imps = BF.openImagePlus(folder + filename) # Open file with Bio-format Plugin
    imp0 = imps[0] #imp0: Original Image 
    if Mode == "Auto":
        bitdepth = imp0.getBitDepth()
        if bitdepth == 8:
            Mode = "8-bit"
        elif bitdepth ==16:
            Mode = "16-bit"
        elif bitdepth ==32:
            Mode = "32-bit"
        else:
            Mode = "16-bit"
    if Show == "Yes":
        imp0.show()
    impsCh = ChannelSplitter.split(imp0) #Channel Splitting of imp0
    n = 0
    for i in impsCh: # it needs to convert Grayscale image
        if Mode == "8-bit":
            ImageConverter(i).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(i).convertToGray16()
        else:
            ImageConverter(i).convertToGray32()
        ip = i.getProcessor()
        n = n + 1
    impR =  impsCh[0]
    impG =  impsCh[1]
    impB =  impsCh[2]
    if Mode <> "8bit":
        IJ.saveAsTiff(impG, folder + filename +"_G.tif")
        IJ.saveAsTiff(impB, folder + filename +"_B.tif")
    if Show == "Yes":
        impR.show()
        impG.show()
        impB.show()
    impR2 = Duplicator().run(impR) # impR2: correcting R image
    width = impR.getWidth()
    iwidth = int(width * 0.015) # define width (15%) of sampling region for calculating mean R-G difference
    if RegionSE[0][1] - RegionSE[0][0] >= 1:
        if SmplXEnd - SmplXStart > 0:
            mean2 = calcMean(impR, impG, SmplXStart, SmplXEnd, nYDivide, yStart, yEnd)
        else:
            mean2 = calcMean(impR, impG, RegionSE[0][0]-iwidth, RegionSE[0][0]-1, nYDivide, yStart, yEnd)
        image_correct(impR, impR2, impG,  RegionSE[0][0], RegionSE[0][1], Mode, BmixRatio, RsubRatio, nYDivide, mean2, yStart, yEnd)
    if RegionSE[1][1] - RegionSE[1][0] >= 1:
        if SmplXEnd - SmplXStart > 0:
            mean2 = calcMean(impR, impG, SmplXStart, SmplXEnd, nYDivide, yStart, yEnd)
        else:
            mean2 = calcMean(impR, impG, RegionSE[1][0]-iwidth, RegionSE[1][0]-1, nYDivide, yStart, yEnd)
        image_correct(impR, impR2, impG,  RegionSE[1][0], RegionSE[1][1], Mode, BmixRatio, RsubRatio, nYDivide, mean2, yStart, yEnd)
    if RegionSE[2][1] - RegionSE[2][0] >= 1:
        if SmplXEnd - SmplXStart > 0:
            mean2 = calcMean(impR, impG, SmplXStart, SmplXEnd, nYDivide, yStart, yEnd)
        else:
            mean2 = calcMean(impR, impG, RegionSE[2][0]-iwidth, RegionSE[2][0]-1, nYDivide, yStart, yEnd)
        image_correct(impR, impR2, impG,  RegionSE[2][0], RegionSE[2][1], Mode, BmixRatio, RsubRatio, nYDivide, mean2, yStart, yEnd)
    impR2.show()
    mergeimp = RGBStackMerge.mergeChannels([impR2, impG, impB, None, None, None, None], True)
#    IJ.saveAsTiff(mergeimp, folder + filename +"_Merge_Stack.tif")
    # convert the composite image to the RGB image
    RGBStackConverter.convertToRGB(mergeimp)
    IJ.saveAsTiff(mergeimp, folder + filename +"_Merge_8bit.tif")
    mergeimp.show()
    if Mode <> "8-bit":
        IJ.saveAsTiff(impR2, folder + filename +"_R2.tif")

    if Show == "Yes":
        imp0.hide()
        impR.hide()
        impR2.hide()
        impG.hide()
        impB.hide()
#        mergeimp.hide()
    imp0.close()
    impR.close()
    impR2.close()
    impG.close()
    impB.close()
#    mergeimp.close()

## Main Program
RegionSE = [[0 for i in range(2)] for j in range(3)] # variable for ROI
meanIntensity = []
od = OpenDialog("Choose a file")
folder = od.getDirectory()
filename = od.getFileName()
path = folder + filename
IJ.log(path)
IJ.log(folder)
IJ.log(filename)
parm = getParameters()
#  input Parameters
yStart = 0
yEnd = 0
if parm is not None:
# if parm is not null object, then...
    Mode, Show, RegionSE, BmixRatio, RsubRatio, nYDivide, SmplXStart, SmplXEnd, Yres = parm
    
#     Mode: Image Format 16-bit or 8-bit
#     Imege: Output Image Type
if Yres== "Yes":
    parm2 = getParameters2(RegionSE, nYDivide, SmplXStart, SmplXEnd)
#  input Parameters
    if parm2 is not None:
    # if parm is not null object, then...
        RegionSE, nYDivide, SmplXStart, SmplXEnd, yStart, yEnd = parm2
if Mode <> "Stop":
    manipulation(Mode, Show, RegionSE, folder, filename, BmixRatio, RsubRatio, nYDivide, SmplXStart, SmplXEnd, yStart, yEnd)
IJ.log("Finish")
#IJ.run("Quit")

