# Red / Cyan Mask MakerRed 8 & 16bit Version
# Ver. 0.01
# (c)2021 Ohnishi, Yasuo
from ij import IJ, ImagePlus 
from ij import plugin
from ij.io import DirectoryChooser
from ij.io import OpenDialog
from ij.gui import GenericDialog
from ij.process import ImageConverter
from ij.process import ImageProcessor
from ij.plugin import ChannelSplitter
from ij.plugin import Duplicator
from ij.plugin.filter import ThresholdToSelection
from loci.plugins import BF
import os
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')

#----------------------------------------------
### function of Value Converting depend on Mode
def convValue(Value, Mode):
#   Converting Luminance Value each in 8, 16, 32bit mode
    if Mode == "16-bit":
        Value = Value * 256
    elif Mode == "32-bit":
        Value = Value * 256 * 256
    else:
        Value = Value
    return Value
#----------------------------------------------
### RGB Manipulation 
def RGBmanipulation(folder,filename,parm):
#  input Parameters
    if parm is not None: # if parm is not null object, then...
        Mode, RTh,  blInvert, baseCh, zoneLTh, zoneHTh,Show = parm # Mode: 16/8bit RTh: Yellow Mask Threshold
      #  Show = "Yes" # temp 
#     Mode: Image Format 16-bit or 8-bit
#     parm_Binary: Threshold of fore / background
#     parm_Distance: Lightness of Distance adjusting Layer
    imps = BF.openImagePlus(folder + filename) # Open file with Bio-format Plugin
    imp0 = imps[0] #imp0: Original Image 
    if Mode == "Auto":
        bitdepth = imp0.getBitDepth()
        print(str(bitdepth))
        if bitdepth == 8:
            Mode = "8-bit"
        elif bitdepth ==16:
            Mode = "16-bit"
        elif bitdepth ==32:
            Mode = "32-bit"
        else:
            Mode = "16-bit"
    print(Mode) 
    if Show == "Yes":
 # Windows Showing Mode
        imp0.show()
    impsCh = ChannelSplitter.split(imp0) #Channel Splitting of imp0
    for i in impsCh: # it needs to convert Grayscale image
        if Mode == "8-bit":
            ImageConverter(i).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(i).convertToGray16()
        else:
            ImageConverter(i).convertToGray32()
## R ch.
    impR =  impsCh[0]
## G ch.
    impG =  impsCh[1]
## B ch. & B x 0.3
    impB =  impsCh[2]
# Show Windows
#    imp0.close()
    if Show == "Yes":
        impR.show()
        impG.show()
        impB.show()
# G ch. x 0.5
    impG05 = Duplicator().run(impG)
    IJ.run(impG05, "Multiply...", "value=0.5")
    if Show == "Yes":
        impG05.show()
# B ch. x 0.5
    impB05 = Duplicator().run(impB)
    IJ.run(impB05, "Multiply...", "value=0.5")
    if Show == "Yes":
        impB05.show()
# R x 0.5 + G x 0.5
    ic = plugin.ImageCalculator()
    impGB = ic.run("Add created", impB05, impG05)
# Close Windows
    if Show == "Yes":
        imp0.hide()
        impR.hide()
        impG.hide()
        impB.hide()
        impGB.show()
        impG05.hide()
        impB05.hide()
    impG05.close()
    impB05.close()
# R - (B x 0.5 + G x 0.5)
    ic = plugin.ImageCalculator()
# Make Red Image Files 
# in each Threshold, Make 3 kinds of Image Files: x1, x2, x4 
    V = str(RTh)
# Red Mask Threshold
    V2 = str(convValue(abs(RTh), Mode)) # actual Yellow mask Threshold
    if blInvert == False:
        # for making Red mask
           # Set the Threshold in B + G Image
        if RTh < 0: # restrict less region for red mask
            print (RTh, "Subtract", V)
            IJ.run(impGB, "Add...","value=" + V2)
            V= "-" + V
        else:  #including more region for red mask
            print (RTh, "Add", V)
            IJ.run(impGB, "Subtract...", "value=" + V2)
            V= "+" + V
        impRmask = ic.run("Subtract created", impR,impGB) # Make Red Mask
        V = "R" + V
    else:
     # for making Cyan mask
         # Set the Threshold in R + G Image
        if RTh >= 0:  #including more region for Cyan mask
            print (RTh, "Add", V)
            IJ.run(impGB, "Add...","value=" + V2)
            V= "+" + V
        else: # restrict less region for cyan mask
            print (RTh, "Subtract", V)
            IJ.run(impGB, "Subtract...", "value=" + V2)
            V= "-" + V
        impRmask = ic.run("Subtract created", impGB, impR) # Make Cyan Mask
        V = "Cy" + V
#       IJ.saveAsTiff(imp6, folder + filename +"_YMask_Thr_"+ V + "_x1.tif") # Save the Subtracted image x 1
    if Show == "Yes":
        impRmask.show()  # Show the subtracted image in window
    IJ.run(impRmask, "Multiply...", "value=4") # making mask more transparent (x 4)
#    if blInvert == True:
#        impRmask.getProcessor().invert()
#        V = V + "Inv"
    IJ.saveAsTiff(impRmask, folder + filename +"_Mask_Thr_"+ V + "_x4.tif") # Save the Subtracted image x 4
#   save Red / Cyan mask
    impGB.hide()
    impGB.close()
#############################################################
# select base channel for Zone mask
    if baseCh == "RGBtoGRAY":
        impZone = Duplicator().run(imp0)
        IJ.run(impZone, "Conversions...", "weighted")
        if Mode == "8-bit":
            ImageConverter(impZone).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(impZone).convertToGray16()
        else:
            ImageConverter(impZone).convertToGray32()
    elif baseCh == "Blue":
        impZone = Duplicator().run(impB)
    elif baseCh == "Red":
        impZone = Duplicator().run(impR)
    elif baseCh == "Green":
        impZone = Duplicator().run(impG)
    impR.close()
    impG.close()
    impB.close()
    imp0.close()
    if Show == "Yes":
        impZone.show()
# make Zone mask
    impZone.getProcessor().setThreshold(convValue(zoneLTh, Mode),convValue(zoneHTh, Mode), ImageProcessor.NO_LUT_UPDATE)
    print "Zone L & H", convValue(zoneLTh, Mode), convValue(zoneHTh, Mode)
    roi = ThresholdToSelection.run(impZone)
    impZone.setRoi(roi)
    try:
        maskimp = ImagePlus("Mask", impZone.getMask()) # only 8bit Zone mask made from ROI: Zone mask: maskimp
    except: # exception process for ROI missing
        IJ.error("Selected ROI (Region of Interest) is missing.")
        if Show == "Yes":
            impRmask.hide()
            impZone.hide()
        impRmask.close()
        impZone.close()
        return
    if Show == "Yes":
        maskimp.show() # show Zone mask
    IJ.saveAsTiff(maskimp, folder + filename +"_L_Zone" + str(int(zoneLTh)) + "-" + str(int(zoneHTh)) + "_Mask.tif")
#        Save Zone Mask
# deepen Zone Mask bit depth  (8bit to 16 or 32 bit)      
    if Mode == "16-bit":
        ImageConverter(maskimp).convertToGray16()
    elif Mode == "32-bit":
        ImageConverter(maskimp).convertToGray32()
    IJ.run(maskimp, "Multiply...", "value=" + str(convValue(1, Mode)))
# bind Zone mask and Yellow / blue mask
    ic = plugin.ImageCalculator()
    maskBindimp = ic.run("Min created", impRmask, maskimp) # bind masks
    if Show == "Yes":
        maskBindimp.show()
    IJ.saveAsTiff(maskBindimp, folder + filename +"_L_Zone" + str(int(zoneLTh)) + "-" + str(int(zoneHTh)) + "_" + V + "_Mask_Binded.tif")
#     save binded mask
    if Show == "Yes":
        impRmask.hide()
        impZone.hide()
        maskimp.hide()
        maskBindimp.hide()
    impRmask.close()
    maskimp.close()
    maskBindimp.close()
    impZone.close()
#   End of def RGBmanipulation 
#----------------------------------------------
### Get parameters from the input DialogBox 
def getParameters():
    gd = GenericDialog("Red (Cyan) Mask Maker / Input parameters")
    mode = ["Auto","16-bit","8-bit"]
    gd.addChoice("Image bit depth", mode, mode[0])
    gd.addNumericField("Red Mask Threshold (-20 to 30)", 10, 0)
    gd.addCheckbox("Mask Invert for Cyan Mask", False)
    sbaseCh = ["RGBtoGRAY","Red","Green","Blue"]
    gd.addChoice("Base Channel for L Zone", sbaseCh, sbaseCh[1])
    gd.addNumericField("Lowest Threshold of L Zone (0 to 255)", 0, 0)
    gd.addNumericField("Highest Threshold of L Zone (0 to 255)", 254, 0)
    showWin = ["Yes","No"]
    gd.addChoice("Show working windows?", showWin, showWin[0])
    gd.showDialog()
    parm_Mode   = gd.getNextChoice()
    RTh =  gd.getNextNumber()
    blInvert = gd.getNextBoolean()
    baseCh =  gd.getNextChoice()
    zoneLTh =  gd.getNextNumber()
    zoneHTh =  gd.getNextNumber()
    Show =  gd.getNextChoice()
    if gd.wasCanceled():  
            parm_Mode   = "Stop"
#            return str(parm_Mode)
    return str(parm_Mode), RTh, blInvert, str(baseCh), zoneLTh, zoneHTh, str(Show)
#----------------------------------------------
# Main Program
#
od = OpenDialog("Choose a file")
folder = od.getDirectory()
file = od.getFileName()
path = folder + file
IJ.log(path)
IJ.log(folder)
IJ.log(file)
parm = getParameters()

if parm[0] <> "Stop":
    RGBmanipulation(folder,file,parm)
IJ.log("Finish")
#IJ.run("Quit")
